from vhh_od.CustObject import CustObject
from vhh_od.utils import append_dict_as_row
import numpy as np


class Shot(object):
    """
    This class is representing a shot. Each instance of this class is holding the properties of one shot.
    """

    def __init__(self, movie_name, sid, start_pos, end_pos):
        """
        Constructor

        :param sid [required]: integer value representing the id of a shot
        :param movie_name [required]: string representing the name of the movie
        :param start_pos [required]: integer value representing the start frame position of the shot
        :param end_pos [required]: integer value representing the end frame position of the shot
        """
        #print("create instance of shot ...");
        self.sid = sid
        self.movie_name = movie_name
        self.start_pos = start_pos
        self.end_pos = end_pos
        self.object_list = []

    def addCustomObject(self, obj: CustObject):
        self.object_list.append(obj)

    def convert2String(self):
        """
        Method to convert class member properties in a semicolon separated string.

        :return: string holding all properties of one shot.
        """
        tmp_str = str(self.sid) + ";" + str(self.movie_name) + ";" + str(self.start_pos) + ";" + str(self.end_pos)
        return tmp_str

    def printShotInfo(self):
        """
        Method to a print summary of shot properties.
        """
        print("------------------------")
        print("shot id: " + str(self.sid))
        print("movie name: " + str(self.movie_name))
        print("start frame: " + str(self.start_pos))
        print("end frame: " + str(self.end_pos))
        print("<<< Object list >>>")
        for obj in self.object_list:
            obj.printObjectInfo()

    def make_end_pos_compatible_with_stc(self):
        """
        End_pos in STC is always included in a shot i.e. end_pos 130 means that the shot contains frame 130
        In OD end_pos does not inlcude the last frame, so end_pos 130 means that the last frame of the shot is 129
        As end_pos gets incriminated when loading STC data, we need to subtract 1 from end_pos to make it compatible with STC.
        """
        self.end_pos -= 1

    def convertObjectList2Dict(self):
        #print("Export objects to csv file ... ")
        dict_l = []
        for idx, obj in enumerate(self.object_list):
            entry_dict = {
                'sid': self.sid,
                'movie_name': self.movie_name,
                'start': self.start_pos,
                'stop': self.end_pos,
                'fid': obj.fid,
                'oid': obj.oid,
                'bb_x1': obj.bb_x1,
                'bb_y1': obj.bb_y1,
                'bb_x2': obj.bb_x2,
                'bb_y2': obj.bb_y2,
                'object_conf': obj.object_conf,
                'class_score': obj.class_score,
                'class_name': obj.object_class_name
            }
            dict_l.append(entry_dict)
        return dict_l

    def update_person_classification(self):
        """ 
        Update the person classification 
        I.e. if the classifier classified a person as a soldier 
        then the class will be changed from "Person" to "Person (Soldier)"
        """
        for obj in self.object_list:
            if obj.person_classification is not None:
                obj.update_with_person_classification()

    def update_obj_classifications(self, use_majority_voting, others_factor = 1):
        """
        Updates the class names of objects according to their classification. 
        If use_majority_voting is True:
            Then objects with the same ID will have the same class (the majority class) in every frame
        """

        if not use_majority_voting:
            self.update_person_classification()
            return

        # Gather all objects according to their tracking id
        objs = {}
        for obj in self.object_list:
            if str(obj.oid) in objs:
                objs[str(obj.oid)].append(obj)
            else:
                objs[str(obj.oid)] = [obj]

        # Do majority voting    
        do_majority_classification(objs)

        # Gather all objects with person classification according to their tracking id
        persons = {}
        for obj in self.object_list:
            if obj.person_classification is None:
                continue

            if str(obj.oid) in persons:
                persons[str(obj.oid)].append(obj)
            else:
                persons[str(obj.oid)] = [obj]

        self.update_person_classification()

        # Do majority voting on the person class  
        do_majority_classification(persons, change_weight_of_others(others_factor))

def do_majority_classification(objs, scale_function = (lambda x,y: y)):
        """
        Set the classification of each object to the majority class.
            obj_list: dictionary where the keys are the object IDs and the values is a list of all CustomObject with this ID
            scale_function: (string, float) -> float 
                Function to weigh the votes for a given class.
                The default value does not change the weight of the votes.
        """
        for obj_list in objs.values():
            # Gather votes
            votes = {}
            for obj in obj_list:
                if obj.object_class_name in votes:
                    votes[obj.object_class_name] += 1
                else:
                    votes[obj.object_class_name] = 1

            # Apply scale_function
            for key in votes.keys():
                votes[key] = scale_function(key, votes[key])

            # Find winner
            winner_idx = np.argmax(votes.values())
            winner = list(votes.keys())[winner_idx]

            # Update the objects class names and classifications according to the winner
            for obj in obj_list:
                obj.update_classification(winner)

def change_weight_of_others(others_factor):
    """
    A function that changes the votes for the "others" 
    """
    def fct(key, votes):
        if key != "person":
            return votes
        else:
            return int(others_factor*votes)
    return fct
                


from vhh_od.utils import *
import yaml


class Configuration:
    """
    This class is needed to read the configuration parameters specified in the configuration.yaml file.
    The instance of the class is holding all parameters during runtime.

    .. note::
       e.g. ./config/config_vhh_test.yaml

        the yaml file is separated in multiple sections
        config['Development']
        config['PreProcessing']
        config['StcCore']
        config['Evaluation']

        whereas each section should hold related and meaningful parameters.
    """

    def __init__(self, config_file: str):
        """
        Constructor

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                                       must be with extension ".yaml"
        """
        printCustom("create instance of configuration ... ", STDOUT_TYPE.INFO)

        if(config_file.split('.')[-1] != "yaml"):
            printCustom("Configuration file must have the extension .yaml!", STDOUT_TYPE.ERROR)

        self.config_file = config_file

        self.debug_flag = -1
        self.stc_results_path = None
        self.save_debug_pkg_flag = -1

        self.batch_size = -1

        self.save_raw_results = -1
        self.path_postfix_raw_results = None
        self.path_prefix_raw_results = None
        self.path_raw_results = None

        self.save_final_results = -1
        self.path_prefix_final_results = None
        self.path_postfix_final_results = None
        self.path_final_results = None

        self.path_videos = None
        self.path_pre_trained_model = None

        self.path_eval_results = None
        self.path_raw_results_eval = None
        self.save_eval_results = -1
        self.path_gt_data = None

        self.model_config_path = None
        self.confidence_threshold = -1
        self.model_class_names_path = None
        self.model_class_selection_path = None

        self.use_deepsort = -1
        self.ds_model_path = None
        self.ds_max_dist = -1
        self.ds_min_conf = -1
        self.ds_nms_max_overlap = -1
        self.ds_max_iou_dist = -1
        self.ds_max_age = -1
        self.ds_num_init = -1

    def loadConfig(self):
        """
        Method to load configurables from the specified configuration file
        """

        fp = open(self.config_file, 'r')
        config = yaml.load(fp, Loader=yaml.BaseLoader)

        developer_config = config['Development']
        pre_processing_config = config['PreProcessing']
        od_core_config = config['OdCore']
        evaluation_config = config['Evaluation']

        # developer_config section
        self.debug_flag = int(developer_config['DEBUG_FLAG'])
        self.stc_results_path = developer_config['STC_RESULTS_PATH']
        self.save_debug_pkg_flag = int(developer_config['SAVE_DEBUG_PKG'])

        # pre-processing section
        self.resize_dim = (int(pre_processing_config['RESIZE_DIM'].split(',')[0]),
                           int(pre_processing_config['RESIZE_DIM'].split(',')[1]))

        # od_core_config section
        self.batch_size = int(od_core_config['BATCH_SIZE'])

        self.path_stc_results = od_core_config['PATH_STC_RESULTS']

        self.save_raw_results = int(od_core_config['SAVE_RAW_RESULTS'])
        self.path_postfix_raw_results = od_core_config['POSTFIX_RAW_RESULTS']
        self.path_prefix_raw_results = od_core_config['PREFIX_RAW_RESULTS']
        self.path_raw_results = od_core_config['PATH_RAW_RESULTS']

        self.save_final_results = int(od_core_config['SAVE_FINAL_RESULTS'])
        self.path_prefix_final_results = od_core_config['PREFIX_FINAL_RESULTS']
        self.path_postfix_final_results = od_core_config['POSTFIX_FINAL_RESULTS']
        self.path_final_results = od_core_config['PATH_FINAL_RESULTS']

        self.path_videos = od_core_config['PATH_VIDEOS']
        self.path_pre_trained_model = od_core_config['PATH_PRETRAINED_MODEL']

        self.model_config_path = od_core_config['MODEL_CONFIG_PATH']

        self.confidence_threshold = float(od_core_config['MODEL_CONF_THRES'])
        self.nms_threshold = float(od_core_config['MODEL_NMS_THRESH'])

        self.model_class_names_path = od_core_config['MODEL_CLASS_NAMES_PATH']
        self.model_class_selection_path = od_core_config['MODEL_CLASS_SELECTION_PATH']

        self.max_frames = int(od_core_config['MAX_FRAMES'])

        self.use_classifier = od_core_config['USE_CLASSIFIER'] == "1"
        self.use_classifier_majority_voting = od_core_config['USE_CLASSIFIER_MAJORITY_VOTING'] == "1"
        self.classifier_model_path = od_core_config['PATH_TO_CLASSIFIER_MODEL']
        self.classifier_model_architecture = od_core_config['CLASSIFIER_MODEL_ARCHITECTURE']

        self.do_normalize_coordinates = od_core_config['DO_NORMALIZE_COORDINATES'] == "1"
        self.others_factor = od_core_config['OTHERS_FACTOR']
        self.shot_types_do_not_run_od = od_core_config['SHOT_TYPES_TO_NOT_RUN_OD']
        
        #DeepSort Parameters
        if od_core_config["USE_DEEPSORT"] == "1":
            self.use_deepsort = True
        else:
            self.use_deepsort = False
        self.ds_model_path = od_core_config["DEEPSORT_MODEL_PATH"]
        self.ds_max_dist = float(od_core_config["DS_MAX_DIST"])
        self.ds_min_conf = float(od_core_config["DS_MIN_CONF"])
        self.ds_nms_max_overlap = float(od_core_config["DS_NMS_MAX_OVERLAP"])
        self.ds_max_iou_dist = float(od_core_config["DS_MAX_IOU_DIST"])
        self.ds_max_age = int(od_core_config["DS_MAX_AGE"])
        self.ds_num_init = int(od_core_config["DS_NUM_INIT"])

        # evaluation section
        self.path_raw_results_eval = evaluation_config['PATH_RAW_RESULTS']
        self.path_eval_results = evaluation_config['PATH_EVAL_RESULTS']
        self.save_eval_results = int(evaluation_config['SAVE_EVAL_RESULTS'])
        self.path_gt_data = evaluation_config['PATH_GT_ANNOTATIONS']



from vhh_od.OD import OD
import os
import sys
import argparse
from vhh_od.Configuration import Configuration
import numpy as np
import cv2

"""
This is file contains a class that can be used to run the object detector on a folder of frames.
Depending on the used parameters this will store the classified crops or can be used to assess the accuracy of the detector
"""

config_file = "./config/config_vhh_od.yaml"

def prepare_paths_list(path_list, n_imgs):
    """
    Selects n_imgs from the list and shuffles list
    """
    paths_chosen = []

    if n_imgs <= 0:
        paths_chosen = path_list

    while len(path_list) > 0 and n_imgs > 0:
        n_imgs -= 1
        i = np.random.randint(0, len(path_list))
        paths_chosen.append(path_list[i])
        del path_list[i]

    np.random.shuffle(paths_chosen)
    return paths_chosen

def store_crops(information, output_folder_path, crop_dict):
    """
    Stores crops and updates information string
    """
    folder_crop_img_path = os.path.join(output_folder_path, crop_dict["class"])
    if not os.path.isdir(folder_crop_img_path):
        os.mkdir(folder_crop_img_path)

    fullpath_crop_img = os.path.join(folder_crop_img_path, crop_dict["name"])
    cv2.imwrite(fullpath_crop_img, crop_dict["cropped_img"])

    information += ','.join([crop_dict["image_file"], crop_dict["name"], crop_dict["class"], crop_dict["x1"], crop_dict["x2"], crop_dict["y1"], crop_dict["y2"]])
    information += '\n'
    return information

def evaluate(crop_dict):
    """
    Opens a window and asks the user whether the classification is correct
    """

    # In case someone touches a wrong key, keep asking for annotations
    need_to_evaluate = True
    while(need_to_evaluate):
        print("\nClass: ", crop_dict["class"])
        sys.stdout.flush()
        cv2.imshow(crop_dict["class"], crop_dict["cropped_img"])
        k = cv2.waitKey(0)

        # There is a bug that waitKey() sometimes gets "ghost" keypresses, if such a keypress occurs call waitKey() again
        while k == 0:
            k = cv2.waitKey(0)

        cv2.destroyAllWindows()

        if k==27:       # Esc key to stop
            need_to_evaluate = False
            break
        elif k==49:     # Keyvalue for 1
            print("Selected: Classification is CORRECT.\n")
            return True
        elif k==48:     # Keyvalue for 0
            print("Selected: Classification is WRONG.\n")
            return False
        else:
            print("Neither 0 or 1 detected, please make sure to not use the numpad.")

def run_on_folder_return_crops(folder_path, output_folder_path):
    """
    Extracts the crops from each image in folder_path and stores them in output_folder_path
    """
    return runOnAllFramesInFolder(folder_path, output_folder_path, True, False)

def run_on_folder_evaluate_model(folder_path, n_imgs=-1, n_annotations=10):
    """
    Runs the object detector on folder_path, classifying at most n_imgs images and manually asks the user if n_annotations crops are correctly classified
    This is then used to compute the accuracy of the model
    If all images are supposed to be used then set n_imgs to <= 0
    """
    return runOnAllFramesInFolder(folder_path, "", False, True, n_imgs, n_annotations)
    
def runOnAllFramesInFolder(folder_path, output_folder_path, do_store_crops, do_evaluate_model, n_imgs=-1, n_annotations=10):
    """
    Runs object detector on images in folder path, at most n_imgs images will be processed
    If all images are supposed to be used then set n_imgs to <= 0
    If do_store_crops is true then the crops will be stored in output_folder_path
    If do_evaluate_model is true: user will be asked to  manually check if n_annotations crops are correctly classified
    This is then used to compute the accuracy of the model
    """

    np.random.seed(0)
    config_instance = Configuration(config_file)
    config_instance.loadConfig()
    od_instance = OD(config_file)

    nr_crops, nr_correct_crops, crops_evaluated = 0, 0, 0

    image_files = [os.path.join(folder_path, path) for path in os.listdir(folder_path)]
    image_files = prepare_paths_list(image_files, n_imgs)
    
    print("Press 1 if the crop and class is correct, 0 if it is wrong. Do not use the numpad. Use escape to quit.")

    # Text for the information file
    information = ""

    for crop_dict in od_instance.iterate_over_images(image_files):
        nr_crops += 1

        # Store crops and update information string
        if do_store_crops:
            information = store_crops(information, output_folder_path, crop_dict)

        # Check if crop is correct
        if do_evaluate_model and  np.random.random() < float(n_annotations) / len(image_files):
            if crops_evaluated >= n_annotations:
                continue

            # Ties are very small and a crop of a tie is hard to indentify as a tie, hence we do not evalute ties
            if crop_dict["class"] == "tie":
                continue

            crops_evaluated += 1
            is_correct = evaluate(crop_dict)

            if is_correct:
                nr_correct_crops += 1

            print("Annotated {0} / {1}".format(crops_evaluated, n_annotations))               

    # Store information string in file        
    if do_store_crops:
        information_file = open(os.path.join(output_folder_path, "crop_information.txt"), "w")
        information_file.write("original_image, crop_image, class_name, x1, x2, y1, y2\n")
        information_file.write(information)
        information_file.close()

    # Output data
    print("Processed {0} / {0} images\nExtracted {1} crops".format(len(image_files), nr_crops))

    if do_evaluate_model:
        print("Manually checked {0} crops, {1} were correcty classified. That is {2}%.".format(crops_evaluated, nr_correct_crops, nr_correct_crops/crops_evaluated * 100))
        return nr_crops, nr_correct_crops, crops_evaluated



import argparse
import os
import time

import numpy as np
import matplotlib.pyplot as plt
import torch
import torch.backends.cudnn as cudnn
import torchvision

from model import Net

parser = argparse.ArgumentParser(description="Train on market1501")
parser.add_argument("--data-dir",default='data',type=str)
parser.add_argument("--no-cuda",action="store_true")
parser.add_argument("--gpu-id",default=0,type=int)
parser.add_argument("--lr",default=0.1, type=float)
parser.add_argument("--interval",'-i',default=20,type=int)
parser.add_argument('--resume', '-r',action='store_true')
args = parser.parse_args()

# device
device = "cuda:{}".format(args.gpu_id) if torch.cuda.is_available() and not args.no_cuda else "cpu"
if torch.cuda.is_available() and not args.no_cuda:
    cudnn.benchmark = True

# data loading
root = args.data_dir
train_dir = os.path.join(root,"train")
test_dir = os.path.join(root,"test")
transform_train = torchvision.transforms.Compose([
    torchvision.transforms.RandomCrop((128,64),padding=4),
    torchvision.transforms.RandomHorizontalFlip(),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
])
transform_test = torchvision.transforms.Compose([
    torchvision.transforms.Resize((128,64)),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
])
trainloader = torch.utils.data.DataLoader(
    torchvision.datasets.ImageFolder(train_dir, transform=transform_train),
    batch_size=64,shuffle=True
)
testloader = torch.utils.data.DataLoader(
    torchvision.datasets.ImageFolder(test_dir, transform=transform_test),
    batch_size=64,shuffle=True
)
num_classes = max(len(trainloader.dataset.classes), len(testloader.dataset.classes))

# net definition
start_epoch = 0
net = Net(num_classes=num_classes)
if args.resume:
    assert os.path.isfile("./checkpoint/ckpt.t7"), "Error: no checkpoint file found!"
    print('Loading from checkpoint/ckpt.t7')
    checkpoint = torch.load("./checkpoint/ckpt.t7")
    # import ipdb; ipdb.set_trace()
    net_dict = checkpoint['net_dict']
    net.load_state_dict(net_dict)
    best_acc = checkpoint['acc']
    start_epoch = checkpoint['epoch']
net.to(device)

# loss and optimizer
criterion = torch.nn.CrossEntropyLoss()
optimizer = torch.optim.SGD(net.parameters(), args.lr, momentum=0.9, weight_decay=5e-4)
best_acc = 0.

# train function for each epoch
def train(epoch):
    print("\nEpoch : %d"%(epoch+1))
    net.train()
    training_loss = 0.
    train_loss = 0.
    correct = 0
    total = 0
    interval = args.interval
    start = time.time()
    for idx, (inputs, labels) in enumerate(trainloader):
        # forward
        inputs,labels = inputs.to(device),labels.to(device)
        outputs = net(inputs)
        loss = criterion(outputs, labels)

        # backward
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()

        # accumurating
        training_loss += loss.item()
        train_loss += loss.item()
        correct += outputs.max(dim=1)[1].eq(labels).sum().item()
        total += labels.size(0)

        # print 
        if (idx+1)%interval == 0:
            end = time.time()
            print("[progress:{:.1f}%]time:{:.2f}s Loss:{:.5f} Correct:{}/{} Acc:{:.3f}%".format(
                100.*(idx+1)/len(trainloader), end-start, training_loss/interval, correct, total, 100.*correct/total
            ))
            training_loss = 0.
            start = time.time()
    
    return train_loss/len(trainloader), 1.- correct/total

def test(epoch):
    global best_acc
    net.eval()
    test_loss = 0.
    correct = 0
    total = 0
    start = time.time()
    with torch.no_grad():
        for idx, (inputs, labels) in enumerate(testloader):
            inputs, labels = inputs.to(device), labels.to(device)
            outputs = net(inputs)
            loss = criterion(outputs, labels)

            test_loss += loss.item()
            correct += outputs.max(dim=1)[1].eq(labels).sum().item()
            total += labels.size(0)
        
        print("Testing ...")
        end = time.time()
        print("[progress:{:.1f}%]time:{:.2f}s Loss:{:.5f} Correct:{}/{} Acc:{:.3f}%".format(
                100.*(idx+1)/len(testloader), end-start, test_loss/len(testloader), correct, total, 100.*correct/total
            ))

    # saving checkpoint
    acc = 100.*correct/total
    if acc > best_acc:
        best_acc = acc
        print("Saving parameters to checkpoint/ckpt.t7")
        checkpoint = {
            'net_dict':net.state_dict(),
            'acc':acc,
            'epoch':epoch,
        }
        if not os.path.isdir('checkpoint'):
            os.mkdir('checkpoint')
        torch.save(checkpoint, './checkpoint/ckpt.t7')

    return test_loss/len(testloader), 1.- correct/total

# plot figure
x_epoch = []
record = {'train_loss':[], 'train_err':[], 'test_loss':[], 'test_err':[]}
fig = plt.figure()
ax0 = fig.add_subplot(121, title="loss")
ax1 = fig.add_subplot(122, title="top1err")
def draw_curve(epoch, train_loss, train_err, test_loss, test_err):
    global record
    record['train_loss'].append(train_loss)
    record['train_err'].append(train_err)
    record['test_loss'].append(test_loss)
    record['test_err'].append(test_err)

    x_epoch.append(epoch)
    ax0.plot(x_epoch, record['train_loss'], 'bo-', label='train')
    ax0.plot(x_epoch, record['test_loss'], 'ro-', label='val')
    ax1.plot(x_epoch, record['train_err'], 'bo-', label='train')
    ax1.plot(x_epoch, record['test_err'], 'ro-', label='val')
    if epoch == 0:
        ax0.legend()
        ax1.legend()
    fig.savefig("train.jpg")

# lr decay
def lr_decay():
    global optimizer
    for params in optimizer.param_groups:
        params['lr'] *= 0.1
        lr = params['lr']
        print("Learning rate adjusted to {}".format(lr))

def main():
    for epoch in range(start_epoch, start_epoch+40):
        train_loss, train_err = train(epoch)
        test_loss, test_err = test(epoch)
        draw_curve(epoch, train_loss, train_err, test_loss, test_err)
        if (epoch+1)%20==0:
            lr_decay()


if __name__ == '__main__':
    main()



from vhh_od.Configuration import Configuration
from vhh_od.Video import Video
from vhh_od.Models import *
from vhh_od.utils import *
from vhh_od.Shot import Shot
from vhh_od.CustObject import CustObject
from vhh_od.visualize import visualize_video
import vhh_od.helpers as Helpers
import vhh_od.Classifier as Classifier
from deep_sort.deep_sort import DeepSort

import numpy as np
import os, sys
import cv2
from matplotlib import cm
import torch
from torch.autograd import Variable
from torch.utils import data
from torchvision import transforms
from collections import namedtuple

Detection_Data = namedtuple('Detection_Data', 'x1 x2 y1 y2 ids obj_class obj_conf class_conf num_results')

class OD(object):
    """
        Main class of shot type classification (stc) package.
    """

    def __init__(self, config_file: str):
        """
        Constructor

        :param config_file: [required] path to configuration file (e.g. PATH_TO/config.yaml)
                                       must be with extension ".yaml"
        """

        if (config_file == ""):
            printCustom("No configuration file specified!", STDOUT_TYPE.ERROR)
            exit()

        self.config_instance = Configuration(config_file)
        self.config_instance.loadConfig()

        if (self.config_instance.debug_flag == True):
            print("DEBUG MODE activated!")
            self.debug_results = "/data/share/maxrecall_vhh_mmsi/develop/videos/results/vhh_od/develop/"

        # prepare object detection model
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.use_tracker = self.config_instance.use_deepsort

        if self.use_tracker:
            printCustom(f"Initializing Deep Sort Tracker...", STDOUT_TYPE.INFO)
            ds_model_path = self.config_instance.ds_model_path
            ds_max_dist = self.config_instance.ds_max_dist
            ds_min_conf = self.config_instance.ds_min_conf
            ds_nms_max_overlap = self.config_instance.ds_nms_max_overlap
            ds_max_iou_dist = self.config_instance.ds_max_iou_dist
            ds_max_age = self.config_instance.ds_max_age
            ds_num_init = self.config_instance.ds_num_init
            use_cuda = (self.device == "cuda")

            self.tracker = DeepSort(ds_model_path, ds_max_dist, ds_min_conf, ds_nms_max_overlap, ds_max_iou_dist, ds_max_age,
                               ds_num_init, use_cuda=torch.cuda.is_available())
            printCustom(f"Deep Sort Tracker initialized successfully!", STDOUT_TYPE.INFO)

        self.num_colors = 10
        self.color_map = cm.get_cmap('gist_rainbow', self.num_colors)

    def iterate_over_images(self, image_files_full_path):
        """
        Iterates over a given list of images applying the object
        detection to each frame. In each iteration it returns one crop.
        """
        self.advanced_init()
        for a, image_file in enumerate(image_files_full_path):
            print("Processed {0} / {1} images".format(a, len(image_files_full_path)), end ="\r")

            img = cv2.imread(image_file)
            img_orig = img

            tensors = torch.unsqueeze(self.preprocess(img), dim = 0)

            predictions_l = self.runModel(model=self.model, tensor_l=tensors, classes=self.classes, class_filter=self.class_selection)
            frame_based_predictions = predictions_l[0]

            # No predictions means no boundary boxes
            if frame_based_predictions is None:
                continue

            im, x, y, w, h = self.rescale_bb(img_orig, frame_based_predictions)
            detection_data = self.detection_data_without_tracking_to_custom_obj(0, x, y, w, h, frame_based_predictions)

            # Switch color channels 
            im = cv2.cvtColor(im, cv2.COLOR_BGR2RGB)

            if len(frame_based_predictions) > 0:
                for i, data in enumerate(detection_data):
                    crop_img = im[data.bb_y1:data.bb_y2, data.bb_x1:data.bb_x2]

                    # Sometimes OD predicts boundary boxes with coordinates < 0, cannot crop those so ignore them
                    if(data.bb_y1 < 0) or (data.bb_y2 < 0) or (data.bb_x1 < 0) or (data.bb_x2 < 0):
                        continue

                    name_crop_img = data.object_class_name + "_" + str(a) + "_" + str(data.oid) + ".png"
                    
                    yield {
                        "cropped_img": crop_img,
                        "class": data.object_class_name,
                        "name": name_crop_img,
                        "image_file": image_file,
                        "x1": str(data.bb_x1),
                        "x2": str(data.bb_x2),
                        "y1": str(data.bb_y1),
                        "y2": str(data.bb_y2),
                    } 

    def advanced_init(self):
        """"
        Creates and loads weight into the mode, loads classes, loads resized dims, and creates preprocess object
        """

        printCustom(f"Initializing Model using \"{self.config_instance.model_config_path}\"...", STDOUT_TYPE.INFO)
        self.model = Darknet(config_path=self.config_instance.model_config_path,
                        img_size=self.config_instance.resize_dim).to(self.device)

        printCustom(f"Loading Weights from \"{self.config_instance.path_pre_trained_model}\"...", STDOUT_TYPE.INFO)
        if self.config_instance.path_pre_trained_model.endswith(".weights"):
            # Load darknet weights
            self.model.load_darknet_weights(self.config_instance.path_pre_trained_model)
        else:
            # Load checkpoint weights
            self.model.load_state_dict(torch.load(self.config_instance.path_pre_trained_model))

        printCustom(f"Loading Class Names from \"{self.config_instance.model_class_names_path}\"... ", STDOUT_TYPE.INFO)
        self.classes = load_classes(self.config_instance.model_class_names_path)

        printCustom(f"Loading Class Selection from \"{self.config_instance.model_class_selection_path}\"... ", STDOUT_TYPE.INFO)
        self.class_selection = load_classes(self.config_instance.model_class_selection_path)
        printCustom(f"Classes of interest: {self.class_selection}", STDOUT_TYPE.INFO)

        # prepare transformation for vhh_od model
        self.preprocess = transforms.Compose([
            transforms.ToPILImage(),
            # transforms.Resize((int(vid_instance.height), vid_instance.width)),
            # transforms.CenterCrop((int(vid_instance.height), int(vid_instance.height))),
            transforms.Resize(self.config_instance.resize_dim),
            # ToGrayScale(),
            transforms.ToTensor()
            # transforms.Normalize((self.config_instance.mean_values[0] / 255.0,
            #                      self.config_instance.mean_values[1] / 255.0,
            #                      self.config_instance.mean_values[2] / 255.0),
            #                     (self.config_instance.std_dev[0] / 255.0,
            #                      self.config_instance.std_dev[1] / 255.0,
            #                      self.config_instance.std_dev[2] / 255.0))
        ])

        self.resized_dim_y = self.config_instance.resize_dim[0]
        self.resized_dim_x = self.config_instance.resize_dim[1]

        if self.config_instance.use_classifier:
            self.classifier = Classifier.Classifier(self.config_instance.classifier_model_architecture, self.config_instance.classifier_model_path, self.device)


    def detection_data_with_tracking_to_custom_obj(self, frame_id, tracking_results):
        """
        Takes prediction data and outputs the detection data as a list of "CustObject"s
        Use this if you are using a tracker
        """
        x1_list = tracking_results[:,0]
        x2_list = tracking_results[:,2]
        y1_list = tracking_results[:,1]
        y2_list = tracking_results[:,3]
        ids = tracking_results[:,4]
        object_classes = tracking_results[:,5]
        object_confs = None
        class_confs = None
        num_results = len(tracking_results)

        data = Detection_Data(x1_list, x2_list, y1_list, y2_list, ids, object_classes, object_confs, class_confs, num_results)
        return self.detection_data_to_custom_obj(data, frame_id)

    def detection_data_without_tracking_to_custom_obj(self, frame_id, x, y, w, h, frame_based_predictions):
        """
        Takes prediction data and outputs the detection data as a list of "CustObject"s
        Use this if you are NOT using a tracker
        """
        x1_list = x
        x2_list = x + w
        y1_list = y
        y2_list = y + h
        ids = None
        object_confs = frame_based_predictions[:,4].cpu().numpy()
        class_confs = frame_based_predictions[:,5].cpu().numpy()
        object_classes = frame_based_predictions[:, 6].cpu().numpy()
        num_results = len(frame_based_predictions)
        data = Detection_Data(x1_list, x2_list, y1_list, y2_list, ids, object_classes, object_confs, class_confs, num_results)
        return self.detection_data_to_custom_obj(data, frame_id)


    def detection_data_to_custom_obj(self, data, frame_id):
        """
        Takes a the detection data in the format of a "Detection_Data" named tuple, outputs a list of "CustObject"s
        """
        obj_id = 0
        cust_object_list = []
        for object_idx in range(data.num_results):
            x1 = int(data.x1[object_idx])
            x2 = int(data.x2[object_idx])
            y1 = int(data.y1[object_idx])
            y2 = int(data.y2[object_idx])

            if data.ids is None:
                instance_id = obj_id
                obj_id = obj_id + 1
            else:
                instance_id = data.ids[object_idx]

            if data.obj_conf is None:
                obj_conf = "N/A"
            else:
                obj_conf = data.obj_conf[object_idx]

            if data.class_conf is None:
                class_conf = "N/A"
            else:
                class_conf = data.class_conf[object_idx]

            class_idx = int(data.obj_class[object_idx])
            class_name = self.classes[class_idx]

            obj_instance = CustObject(oid=instance_id,
                                        fid=frame_id,
                                        object_class_name=class_name,
                                        object_class_idx=class_idx, 
                                        object_conf=obj_conf,
                                        class_score=class_conf,
                                        bb_x1=x1,
                                        bb_y1=y1,
                                        bb_x2=x2,
                                        bb_y2=y2
                                        )
            cust_object_list.append(obj_instance)
        return cust_object_list

    def rescale_bb(self, image_orig, frame_based_predictions):
        """
        Rescales bounding boxes to fit original video resolution
        """
        im = cv2.cvtColor(image_orig, cv2.COLOR_BGR2RGB)
        y_factor = im.shape[0] / self.resized_dim_y
        x_factor = im.shape[1] / self.resized_dim_x
        x = (frame_based_predictions[:, 0]).cpu().numpy() * x_factor
        y = (frame_based_predictions[:, 1]).cpu().numpy() * y_factor
        w = (frame_based_predictions[:, 2]).cpu().numpy() * x_factor - x
        h = (frame_based_predictions[:, 3]).cpu().numpy() * y_factor - y
        return im, x, y, w, h

    def apply_tracker(self, x, y, w, h, im, frame_id, new_custom_objects, frame_based_predictions, vis = False):
         # Convert BBoxes from XYXY (corner points) to XYWH (center + width/height) representation
        x = x+w/2
        y = y+h/2
        bbox_xywh = np.array([[x[i],y[i],w[i],h[i]] for i in range(len(frame_based_predictions))])

        # get class confidences
        cls_conf = frame_based_predictions[:, 5].cpu().numpy()
        class_predictions = frame_based_predictions[:, 6].cpu().numpy()

        # Track Objects using Deep Sort tracker
        # Tracker expects Input as XYWH but returns Boxes as XYXY
        tracking_results = np.array(self.tracker.update(bbox_xywh, cls_conf, class_predictions, im))
        num_results = len(tracking_results)

        if num_results > 0:
            detection_data = self.detection_data_with_tracking_to_custom_obj(frame_id, tracking_results)
        else:
            # Since we need to advance the framecounter of the classifier, we keep track of frames with no predictions
            new_custom_objects.append(None)
            detection_data = []

        if not vis:
            return detection_data

        # visualization
        num_colors = 10
        color_map = cm.get_cmap('gist_rainbow', num_colors)
        if len(tracking_results) > 0:
            for box in tracking_results:
                x1v = int(box[0])
                x2v = int(box[2])
                y1v = int(box[1])
                y2v = int(box[3])
                color_idx = box[4] % self.num_colors
                color = color_map(color_idx)[0:3]
                color = tuple([int(color[i] * 255) for i in range(len(color))])
                class_name = self.classes[int(box[5])]
                label = f"{class_name} {box[4]}"
                font_size = 0.5
                font_thickness = 1
                text_size = cv2.getTextSize(label, cv2.FONT_HERSHEY_SIMPLEX, font_size , font_thickness)[0]
                # draw bounding box
                im = cv2.rectangle(im, (x1v, y1v), (x2v, y2v), color, 5)
                # draw text and background
                cv2.rectangle(im, (x1v, y1v), (x1v + text_size[0] + 3, y1v + text_size[1] + 4), color, -1)
                cv2.putText(im, label, (x1v, y1v + text_size[1]), cv2.FONT_HERSHEY_SIMPLEX, font_size,
                            [0, 0, 0], font_thickness)
            cv2.imshow("im", im)
            cv2.waitKey()
        return detection_data

    def get_video_instance(self, shots_per_vid_np, max_recall_id):
        """
        Generate the video instance on which object detection will be run
        Also handles initialization and checking of parameters for runOnSingleVideo(...)
        """
        if (type(shots_per_vid_np) == None):
            print("ERROR: you have to set the parameter shots_per_vid_np!")
            exit()

        if (max_recall_id == -1 or max_recall_id == 0):
            print("ERROR: you have to set a valid max_recall_id [1-n]!")
            exit()

        if(self.config_instance.debug_flag == True):
            # load shot list from result file
            printCustom(f"Loading STC Results from \"{self.config_instance.path_stc_results}\"...", STDOUT_TYPE.INFO)
            shots_np = self.loadStcResults(self.config_instance.path_stc_results)
        else:
            shots_np = shots_per_vid_np

        if (len(shots_np) == 0):
            print("ERROR: there must be at least one shot in the list!")
            exit()

        if (self.config_instance.debug_flag == True):
            num_shots = 10
            offset = 0
        else:
            num_shots = len(shots_per_vid_np)
            offset = 0

        # Ensure tracker is empty
        if self.use_tracker:
            self.tracker.tracker.clear_id()
            self.tracker.reset()

        # load video instance
        vid_name = shots_np[0][0]
        vid_instance = Video()
        vid_instance.load(os.path.join(self.config_instance.path_videos, vid_name))

        # prepare numpy shot list
        shot_instance = None
        for s in range(offset, offset + num_shots):
            # print(shots_per_vid_np[s])

            # Do not use shots that have the wrong shot type
            if shots_per_vid_np[s][4] in self.config_instance.shot_types_do_not_run_od:
                continue

            shot_instance = Shot(sid=int(s + 1),
                                 movie_name=shots_per_vid_np[s][0],
                                 start_pos=int(shots_per_vid_np[s][2]),
                                 end_pos=int(shots_per_vid_np[s][3]) + 1 )

            vid_instance.addShotObject(shot_obj=shot_instance)

        self.advanced_init()
        return vid_instance

    def process_predictions(self, predictions_l, images_orig, frame_id, results_od_l, new_custom_objects, current_shot):
        shot_id = int(current_shot.sid)
        vid_name = str(current_shot.movie_name)
        start = int(current_shot.start_pos)
        stop = int(current_shot.end_pos) 

        # For each frame, track predictions and store results
        for a in range(0, len(predictions_l)):
            frame_id += 1
            frame_based_predictions = predictions_l[a]
            detection_data = []

        
            if (frame_based_predictions is None):
                results_od_l.append(["None", shot_id, vid_name, start, stop, frame_id,
                                        "None", "None", "None", "None", "None", "None", "None"])

                if (self.config_instance.debug_flag == True):
                    tmp = str(None) + ";" + str(shot_id) + ";" + str(vid_name) + ";" + str(start) + ";" + str(
                        stop) + ";" + str(frame_id) + ";" + str(None) + ";" + str(None) + ";" + str(
                        None) + ";" + str(None) + ";" + str(None) + ";" + str(None) + ";" + str(None)
                    print(tmp)

                # Since we need to advance the framecounter of the classifier, we keep track of frames with no predictions
                new_custom_objects.append(None)
                continue

            # rescale bounding boxes to fit original video resolution
            im, x, y, w, h = self.rescale_bb(images_orig[a], frame_based_predictions)

            if self.use_tracker:
                detection_data =  self.apply_tracker(x, y, w, h, im, frame_id, new_custom_objects, frame_based_predictions)
            else:
                # if no tracker is used, store the detection results
                detection_data = self.detection_data_without_tracking_to_custom_obj(frame_id, x, y, w, h, frame_based_predictions)

            # store predictions for each object in the frame
            for obj in detection_data:   
                results_od_l.append([obj.oid, shot_id, vid_name, start, stop, frame_id,
                                        obj.bb_x1, obj.bb_y1, obj.bb_x2, obj.bb_y2, obj.object_conf, obj.class_score, obj.object_class_idx])  

                current_shot.addCustomObject(obj)


                if (self.config_instance.debug_flag == True):
                    print(obj.printObjectInfo())
            new_custom_objects += detection_data
        return frame_id

    def normalize_bb(self, bb_list, width, height):
        for obj in bb_list:
            obj.bb_x1 /= width
            obj.bb_x2 /= width
            obj.bb_y1 /= height
            obj.bb_y2 /= height

    def runOnSingleVideo(self, shots_per_vid_np=None, max_recall_id=-1):
        """
        Method to run stc classification on specified video.

        :param shots_per_vid_np: [required] numpy array representing all detected shots in a video
                                 (e.g. sid | movie_name | start | end )
        :param max_recall_id: [required] integer value holding unique video id from VHH MMSI system
        """
        print("run vhh_od detector on single video ... ")
        vid_instance = self.get_video_instance(shots_per_vid_np, max_recall_id)

        printCustom(f"Starting Object Detection (Executing on device {self.device})... ", STDOUT_TYPE.INFO)
        results_od_l = []
        previous_shot_id, frame_id = -1, -1

        last_shot_to_process = max([shot.sid for shot in vid_instance.shot_list])
        height, width = None, None
        for shot_frames in vid_instance.getFramesByShots_NEW(preprocess_pytorch=self.preprocess, max_frames_per_return=self.config_instance.max_frames):
            shot_tensors, images_orig, current_shot = shot_frames["Tensors"], shot_frames["Images"], shot_frames["ShotInfo"]

            if height is None:
                height, width, _ = images_orig[0].shape

            shot_id, vid_name = int(current_shot.sid), str(current_shot.movie_name)
            start, stop = int(current_shot.start_pos), int(current_shot.end_pos)

            # Collect all custom objects so we can run the classifier on them
            new_custom_objects = []

            if shot_id != previous_shot_id:
                frame_id = start - 1

            # Only run the model if we actually have tensors
            if shot_tensors is None:
                previous_shot_id = shot_id
                continue

            print("{0} / {1} shots".format(shot_id, last_shot_to_process), end="\r")

            if(self.config_instance.debug_flag == True):
                print("-----")
                print(f"Video Name: {vid_name}")
                print(f"Shot ID: {shot_id}")
                print(f"Start: {start} / Stop: {stop}")
                print(f"Duration: {stop - start} Frames")

            # Run vhh_od detector get predictions
            predictions_l = self.runModel(model=self.model, tensor_l=shot_tensors, classes=self.classes, class_filter=self.class_selection)

            # Reset tracker for every new shot
            if self.use_tracker and previous_shot_id != shot_id:
                self.tracker.reset()

            # Process Yolo's predictions and update: results_od_l, new_custom_objects, current_shot
            # This also runs the tracker
            frame_id = self.process_predictions(predictions_l, images_orig, frame_id, results_od_l, new_custom_objects, current_shot)

            # Use classifier on crops
            if self.config_instance.use_classifier:
                Classifier.run_classifier_on_list_of_custom_objects(self.classifier, new_custom_objects, shot_frames["Images"])

            # Add classifier results, do majority voting
            current_shot.update_obj_classifications(self.config_instance.use_classifier_majority_voting, self.config_instance.others_factor)

            # Normalize coordinates
            if self.config_instance.do_normalize_coordinates:
                self.normalize_bb(current_shot.object_list, width, height)

            previous_shot_id = shot_id
        
        if (self.config_instance.debug_flag):
            vid_instance.printVIDInfo()

        # Ensure that the frame window in the output is compatible with STC
        for shot in vid_instance.shot_list:
            shot.make_end_pos_compatible_with_stc()

        # Store results
        if (self.config_instance.save_final_results):
            results_path = Helpers.mkdir_if_necessary(self.config_instance.path_final_results)
            filepath = f"{results_path}{vid_name.split('.')[0]}.{self.config_instance.path_postfix_final_results}"
            vid_instance.export2csv(filepath=filepath)

            if (self.config_instance.save_raw_results):
                results_path = Helpers.mkdir_if_necessary(self.config_instance.path_raw_results)
                visualize_video(vid_instance, filepath, results_path)

                '''
                for shot in vid_instance.shot_list:
                    vid_instance.visualizeShotsWithBB(path=results_path,
                                                    sid=int(shot.sid),
                                                    all_frames_tensors=all_tensors_l,
                                                    save_single_plots_flag=True,
                                                    plot_flag=False,
                                                    boundingbox_flag=True,
                                                    save_as_video_flag=True
                                                    ) '''

    def runModel(self, model, tensor_l, classes, class_filter):
        """
        Method to calculate stc predictions of specified model and given list of tensor images (pytorch).

        :param model: [required] pytorch model instance
        :param tensor_l: [required] list of tensors representing a list of frames.
        :return: predicted class_name for each tensor frame,
                 the number of hits within a shot,
                 frame-based predictions for a whole shot
        """
        # run vhh_od detector

        # prepare pytorch dataloader
        Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor
        dataset = data.TensorDataset(tensor_l)  # create your datset
        inference_dataloader = data.DataLoader(dataset=dataset,
                                               batch_size=self.config_instance.batch_size)

        predictions_l = []
        for i, inputs in enumerate(inference_dataloader):
            input_batch = inputs[0]
            input_batch = Variable(input_batch.type(Tensor))

            # move the input and model to GPU for speed if available
            if torch.cuda.is_available():
                input_batch = input_batch.to('cuda')
                model.to('cuda')

            model.eval()
            with torch.no_grad():
                output = model(input_batch)
                batch_detections = non_max_suppression(prediction=output,
                                                 conf_thres=self.config_instance.confidence_threshold,
                                                 nms_thres=self.config_instance.nms_threshold)

                for frame_detection in batch_detections:

                    filtered_detection = None

                    if frame_detection is not None:

                        for i in range(len(frame_detection)):

                            detected_object = frame_detection[i]
                            class_idx = detected_object[6].int().item()

                            if classes[class_idx] in class_filter:
                                if filtered_detection is None:
                                    filtered_detection = detected_object.unsqueeze(dim=0)
                                else:
                                    filtered_detection = torch.cat([filtered_detection, detected_object.unsqueeze(dim=0)], dim=0)

                    predictions_l.append(filtered_detection)

        return predictions_l

    def loadStcResults(self, stc_results_path):
        """
        Method for loading shot type classification results as numpy array
        :param stc_results_path: [required] path to results file of shot type classification module (vhh_stc)
        :return: numpy array holding list of detected shots and shot types.
        """

        # open sbd results
        fp = open(stc_results_path, 'r')
        lines = fp.readlines()
        lines = lines[1:]

        lines_n = []
        for i in range(0, len(lines)):
            line = lines[i].replace('\n', '')
            line_split = line.split(';')
            lines_n.append([line_split[0], line_split[1], line_split[2], line_split[3], line_split[4]])
        lines_np = np.array(lines_n)

        return lines_np



import logging
import os
from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

# read requirements
install_requires=[]
with open("requirements.txt", "r") as f:
    reqs = f.readlines()
    for i in range(len(reqs)):
        req = reqs[i]
        if i < len(reqs)-1:
            req = req[:-1]
        if req[0] is not '#':
            install_requires.append(req)

# install vhh_sbd package
setup(
     name='vhh_od',
     version='1.4.0',
     author="Daniel Helm",
     author_email="daniel.helm@tuwien.ac.at",
     description="Object Detection and Tracking Package",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/dahe-cvl/vhh_od",
     packages=find_packages(),
     install_requires=install_requires
)
